import logging
from typing import List, Optional, Dict
from rich.console import Console
from rich.panel import Panel
from rich.live import Live
from rich.control import Control
from rich.table import Table
import json
import os
from dotenv import load_dotenv

from langchain_core.language_models import BaseChatModel
from langchain_core.messages import (
    HumanMessage,
    BaseMessage,
    SystemMessage,
    ToolMessage,
)
from langchain_anthropic import ChatAnthropic

from .client import Client

logging.basicConfig(
    level=logging.WARNING, format="%(asctime)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger(__name__)


class Agent:
    """Agent for interacting with MCP tools using LLMs."""

    def __init__(
        self,
        client: Client,
        mode: str = "cli",
        model: str = "claude-3-5-sonnet-20240620",
        system_prompt: Optional[str] = None,
    ):
        """Initialize the Agent.

        Args:
            client: MCP Hive Client instance
            mode: Operation mode ('cli' or 'server')
            model: LLM model to use
            system_prompt: Custom system prompt
        """
        self.client = client
        self.mode = mode
        self.model_name = model
        self.console = Console()

        # Load environment variables
        load_dotenv()

        # Initialize LLM
        self.llm: BaseChatModel = ChatAnthropic(
            model=self.model_name,
            temperature=0.2,
            max_tokens=8192,
        )

        # Initialize chat history with system prompt
        self.system_prompt = system_prompt or (
            "You are an expert assistant. Your goal is to help the user with their query."
        )
        self.chat_history: List[BaseMessage] = [
            SystemMessage(content=self.system_prompt)
        ]

    def _validate_environment(self) -> bool:
        """Validate all required environment variables are set.

        Returns:
            bool: True if all required variables are set, False otherwise
        """
        missing_vars: Dict[str, str] = {}

        # Check Anthropic API key (LLM-specific requirement)
        if not os.getenv("ANTHROPIC_API_KEY"):
            missing_vars["ANTHROPIC_API_KEY"] = "Required for Claude LLM access"

        # Get required variables from client (tool-specific requirements)
        client_env_vars = self.client.get_required_env_vars()
        for var_name, description in client_env_vars.items():
            if not os.getenv(var_name):
                missing_vars[var_name] = description

        if missing_vars:
            self._print_missing_env_vars(missing_vars)
            return False
        return True

    def _print_missing_env_vars(self, missing_vars: Dict[str, str]) -> None:
        """Print a formatted message about missing environment variables.

        Args:
            missing_vars: Dictionary of missing variable names and their descriptions
        """
        self.console.print("\n[bold red]⚠️  Missing Required Environment Variables[/]\n")

        table = Table(show_header=True, header_style="bold cyan")
        table.add_column("Variable Name")
        table.add_column("Description")

        for var_name, description in missing_vars.items():
            table.add_row(f"[yellow]{var_name}[/]", description)

        self.console.print(table)

        self.console.print(
            "\n[bold green]Please create or update your .env file with the following format:[/]\n"
        )

        env_example = "\n".join(
            f"{var_name}=your_{var_name.lower()}_here" for var_name in missing_vars
        )

        self.console.print(
            Panel(
                env_example,
                title="[bold blue].env file[/]",
                border_style="blue",
                padding=(1, 2),
            )
        )

        self.console.print(
            "\n[italic]Make sure to keep your .env file secure and never commit it to version control.[/]\n"
        )

    def configure_behavior(self, system_prompt: str) -> None:
        """Update the agent's system prompt.

        Args:
            system_prompt: New system prompt for the agent
        """
        self.system_prompt = system_prompt
        self.chat_history = [SystemMessage(content=system_prompt)]

    @staticmethod
    def _print_banner():
        """Print the MCP HIVE banner."""
        banner = """
\033[3m\033[38;2;66;139;202m\
███╗   ███╗ ██████╗██████╗     ██╗  ██╗██╗██╗   ██╗███████╗
████╗ ████║██╔════╝██╔══██╗    ██║  ██║██║██║   ██║██╔════╝
██╔████╔██║██║     ██████╔╝    ███████║██║██║   ██║█████╗  
██║╚██╔╝██║██║     ██╔═══╝     ██╔══██║██║╚██╗ ██╔╝██╔══╝  
██║ ╚═╝ ██║╚██████╗██║         ██║  ██║██║ ╚████╔╝ ███████╗
╚═╝     ╚═╝ ╚═════╝╚═╝         ╚═╝  ╚═╝╚═╝  ╚═══╝  ╚══════╝\033[0m
                                                            
\033[3m\033[38;2;114;159;207m        🐝 The hub for all your LLM tools 🐝\033[0m
        """
        print(banner)

    async def process_query(self, query: str) -> None:
        """Process a user query using LLM and available tools.

        Args:
            query: User's input query
        """
        self.chat_history.append(HumanMessage(content=query))

        # Create a Live display context for real-time updates
        with Live("", auto_refresh=True, console=self.console) as live:
            complete_response = []
            while True:
                if len(complete_response) == 0:
                    complete_response.append("[dim]Thinking...[/]")
                else:
                    complete_response.append("[dim]Processing tool results...[/]")

                panel = Panel(
                    "\n".join(complete_response),
                    title="[bold purple]🤖 Assistant[/]",
                    border_style="purple",
                    padding=(1, 2),
                )
                live.update(panel)

                # Get LLM response
                ai_msg = await self.llm.ainvoke(self.chat_history)
                self.chat_history.append(ai_msg)

                # Handle regular message content
                if ai_msg.content:
                    content = (
                        ai_msg.content[0].get("text", "")
                        if isinstance(ai_msg.content, list)
                        else ai_msg.content
                    )
                    complete_response[-1] = content

                # Process tool calls
                for tool_call in ai_msg.tool_calls:
                    tool_args = json.dumps(tool_call["args"], indent=2)

                    # Show tool call progress
                    tool_section = (
                        f"\n🔧 [yellow]Tool:[/] {tool_call['name']}\n"
                        f"[dim]Arguments:[/]\n```\n{tool_args}\n```\n"
                        f"[dim]Running tool [green]{tool_call['name']}...[/]"
                    )
                    complete_response.append(tool_section)
                    live.update(
                        Panel(
                            "\n".join(complete_response),
                            title="[bold purple]🤖 Assistant[/]",
                            border_style="purple",
                            padding=(1, 2),
                        )
                    )

                    # Execute tool call
                    try:
                        result = await self.client.call_tool(
                            tool_call["name"], tool_call["args"]
                        )
                        logger.info(f"Tool call result: {result}")
                        tool_msg = ToolMessage(
                            content=result.content,
                            name=tool_call["name"],
                            tool_call_id=tool_call["id"],
                        )
                        success = not (
                            result.content[0].type == "text"
                            and result.content[0].text.startswith(
                                "Error executing tool"
                            )
                        )
                    except Exception as e:
                        logger.error(f"Error calling tool {tool_call['name']}: {e}")
                        tool_msg = ToolMessage(
                            content=f"Error calling tool {tool_call['name']}: {str(e)}",
                            name=tool_call["name"],
                            tool_call_id=tool_call["id"],
                        )
                        success = False

                    self.chat_history.append(tool_msg)

                    # Update tool call status
                    status_message = (
                        "[green]✓ Succeeded[/]" if success else "[red]✗ Failed[/]"
                    )
                    tool_section = (
                        f"\n🔧 [yellow]Tool:[/] {tool_call['name']}\n"
                        f"[dim]Arguments:[/]\n```\n{tool_args}\n```\n"
                        f"{status_message}"
                    )
                    complete_response[-1] = tool_section
                    live.update(
                        Panel(
                            "\n".join(complete_response),
                            title="[bold purple]🤖 Assistant[/]",
                            border_style="purple",
                            padding=(1, 2),
                        )
                    )

                if len(ai_msg.tool_calls) == 0:
                    # If no tool calls, display message and break
                    live.update(
                        Panel(
                            "\n".join(complete_response),
                            title="[bold purple]🤖 Assistant[/]",
                            border_style="purple",
                            padding=(1, 2),
                        )
                    )
                    break

    async def cli_chat_loop(self) -> None:
        """Run an interactive CLI chat loop."""
        self._print_banner()

        self.console.print(
            Panel(
                "[bold green]Welcome to the Interactive Chat![/]\n"
                "[italic]Type 'exit' or 'quit' to end the conversation[/]",
                title="[bold blue]Chat Started[/]",
                border_style="blue",
            )
        )

        while True:
            try:
                query = self.console.input("\n[bold blue]You 💭:[/] ").strip()
                if query.lower() in ["quit", "exit"]:
                    self.console.print("\n[bold red]Goodbye! 👋[/]")
                    break

                # Move cursor up and remove input prompt
                self.console.print(Control.move(0, -1), end="")

                # Display user message
                self.console.print(
                    Panel(
                        query,
                        title="[bold blue]You[/]",
                        border_style="blue",
                        padding=(1, 2),
                    )
                )

                await self.process_query(query)
            except Exception as e:
                logger.error(f"Error in chat loop: {e}")
                self.console.print(f"[bold red]Error:[/] {str(e)}")

    async def run(self) -> None:
        """Run the agent in the specified mode."""
        try:
            # Connect to hives and set up tools only if validation passes
            await self.client.connect()

            # Validate environment variables first
            if not self._validate_environment():
                return

            # Bind tools to LLM
            llm_tools = [tool.to_llm_tool() for tool in self.client.tools]
            self.llm = self.llm.bind_tools(llm_tools)

            if self.mode == "cli":
                await self.cli_chat_loop()
            elif self.mode == "server":
                raise NotImplementedError("Server mode not yet implemented")
            else:
                raise ValueError(f"Unknown mode: {self.mode}")
        finally:
            await self.client.cleanup()


def create_agent(
    hives: List[str],
    mode: str = "cli",
    model: str = "claude-3-5-sonnet-20240620",
    system_prompt: Optional[str] = None,
) -> Agent:
    """Create a new MCP Agent with the specified configuration.

    Args:
        hives: List of MCP hive URLs
        mode: Operation mode ('cli' or 'server')
        model: LLM model to use
        system_prompt: Custom system prompt

    Returns:
        Configured Agent instance
    """
    client = Client(hives=hives)
    return Agent(
        client=client,
        mode=mode,
        model=model,
        system_prompt=system_prompt,
    )



import json
import logging
from typing import Dict, List, Any
from contextlib import AsyncExitStack
from dotenv import load_dotenv
import os

from mcp.client.sse import sse_client
from mcp.client.session import ClientSession
from mcp.types import CallToolResult

from .types import Tool, ToolConfig, SecurityScheme, ApiKeySecurityScheme

logger = logging.getLogger(__name__)
load_dotenv()


class Client:
    """Client for managing MCP server connections and tools."""

    def __init__(self, hives: List[str]):
        self.hives = hives
        self.sessions: Dict[str, ClientSession] = {}
        self.exit_stack = AsyncExitStack()
        self.tools: List[Tool] = []
        self._tools_by_name: Dict[str, Tool] = {}

    async def connect(self) -> None:
        """Connect to all configured hives."""
        for server_url in self.hives:
            try:
                sse_transport = await self.exit_stack.enter_async_context(
                    sse_client(url=server_url)
                )
                stdio, write = sse_transport
                session = await self.exit_stack.enter_async_context(
                    ClientSession(stdio, write)
                )
                await session.initialize()
                self.sessions[server_url] = session
                logger.info(f"Connected to hive: {server_url}")
            except Exception as e:
                logger.error(f"Failed to connect to {server_url}: {e}")

        await self.setup_tools()

    async def setup_tools(self) -> List[Tool]:
        """Get and filter tools from all connected hives."""
        self.tools = []
        for server_url, session in self.sessions.items():
            try:
                response = await session.list_tools()
                tool_configs = await self._list_tool_configs(session, server_url)
                tool_configs_by_name = {
                    tool_config.name: tool_config for tool_config in tool_configs
                }
                params_to_exclude = await self._get_params_to_exclude(
                    session, server_url
                )

                filtered_tools = [
                    Tool(
                        name=tool.name,
                        description=tool.description,
                        input_schema=self._filter_tool_params(
                            tool.inputSchema, params_to_exclude
                        ),
                        hive=server_url,
                        security_schemes=[
                            SecurityScheme.from_dict(scheme)
                            for scheme in tool_configs_by_name[
                                tool.name
                            ].security_schemes
                        ],
                        host=tool_configs_by_name[tool.name].host,
                    )
                    for tool in response.tools
                    if tool.name not in {"list_tool_configs", "list_params_to_exclude"}
                ]
                self.tools.extend(filtered_tools)
            except Exception as e:
                logger.error(f"Error getting tools from {server_url}: {e}")

        self._tools_by_name = {tool.name: tool for tool in self.tools}
        return self.tools

    async def call_tool(
        self, tool_name: str, tool_args: Dict[str, Any]
    ) -> CallToolResult:
        """Call a tool with the given arguments."""
        tool = self._tools_by_name.get(tool_name)
        if not tool:
            raise ValueError(f"Tool {tool_name} not found")

        session = self.sessions[tool.hive]
        args = tool_args.copy()

        # Add host and auth information
        args["host__"] = os.environ.get(tool.host.get_env_var(), tool.host.default_url)

        if tool.security_schemes:
            security_scheme = tool.security_schemes[0]
            if isinstance(security_scheme, ApiKeySecurityScheme):
                security_scheme_data = security_scheme.to_dict()
                security_scheme_data["value"] = os.environ.get(
                    security_scheme.env_variable, "xxx"
                )
            security_scheme_data["key"] = security_scheme.name
            args["request_auth__"] = security_scheme_data

        return await session.call_tool(tool_name, args)

    async def cleanup(self) -> None:
        """Clean up all sessions."""
        await self.exit_stack.aclose()

    # Helper methods
    async def _list_tool_configs(
        self, session: ClientSession, server_url: str
    ) -> List[ToolConfig]:
        try:
            result = await session.call_tool("list_tool_configs", {})
            return (
                [ToolConfig(**json.loads(tool.text)) for tool in result.content]
                if result.content
                else []
            )
        except Exception as e:
            logger.warning(f"Could not get tool configs from {server_url}: {e}")
            return []

    async def _get_params_to_exclude(
        self, session: ClientSession, server_url: str
    ) -> List[str]:
        try:
            result = await session.call_tool("list_params_to_exclude", {})
            return [param.text for param in result.content] if result.content else []
        except Exception as e:
            logger.warning(f"Could not get params to exclude from {server_url}: {e}")
            return []

    @staticmethod
    def _filter_tool_params(schema: dict, params_to_exclude: List[str]) -> dict:
        filtered_schema = schema.copy()
        if "properties" in filtered_schema:
            filtered_schema["properties"] = {
                k: v
                for k, v in filtered_schema["properties"].items()
                if k not in params_to_exclude
            }
        return filtered_schema

    @staticmethod
    def _get_host_env_var_name(host_name: str) -> str:
        return f"{host_name.upper()}_SERVER_URL"

    @staticmethod
    def _get_security_env_var_name(host_name: str) -> str:
        return f"{host_name.upper()}_API_KEY"

    def get_required_env_vars(self) -> Dict[str, str]:
        """Get all required environment variables for all tools.

        Returns:
            Dict mapping environment variable names to their descriptions
        """
        env_vars = {}

        # Collect env vars from all tools
        for tool in self.tools:
            tool_env_vars = tool.get_required_env_vars()
            # Update without overwriting existing descriptions
            for var_name, description in tool_env_vars.items():
                if var_name not in env_vars:
                    env_vars[var_name] = description

        return env_vars



from dataclasses import dataclass
from typing import Dict, List, Any
from abc import ABC, abstractmethod


@dataclass
class Host:
    name: str
    default_url: str

    def get_env_var(self) -> str:
        return f"{self.name.upper()}_URL"


@dataclass
class SecurityScheme(ABC):
    @classmethod
    @abstractmethod
    def from_dict(cls, json_data) -> "SecurityScheme":
        if json_data.get("type") == "apiKey":
            return ApiKeySecurityScheme.from_dict(json_data)
        raise ValueError(f"Unsupported security scheme type: {json_data.get('type')}")

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        pass


@dataclass
class ApiKeySecurityScheme(SecurityScheme):
    id: str
    type: str
    name: str
    in_field: str
    env_variable: str

    @classmethod
    def from_dict(cls, json_data) -> "ApiKeySecurityScheme":
        data = json_data.copy()
        data["type"] = "apiKey"
        data["in_field"] = data["in"]
        data["env_variable"] = data["envVariable"]
        del data["in"]
        del data["envVariable"]
        return cls(**data)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "type": self.type,
            "name": self.name,
            "in": self.in_field,
        }


@dataclass
class Tool:
    name: str
    description: str
    input_schema: dict
    hive: str
    security_schemes: List[SecurityScheme]
    host: Host

    def get_required_env_vars(self) -> Dict[str, str]:
        """Get all required environment variables for this tool.

        Returns:
            Dict mapping environment variable names to their descriptions
        """
        env_vars = {}

        env_vars[self.host.get_env_var()] = f"Server URL for {self.host.name}"

        # Add security scheme environment variables
        for scheme in self.security_schemes:
            if scheme.type == "apiKey":
                env_vars[scheme.env_variable] = f"API key for {self.host.name}"

        return env_vars

    def to_llm_tool(self):
        return {
            "name": self.name,
            "description": self.description,
            "input_schema": self.input_schema,
        }


@dataclass
class ToolConfig:
    id: str
    name: str
    security_schemes: List[Dict[str, Any]]
    host: Host

    def __init__(self, **kwargs):
        self.id = kwargs["id"]
        self.name = kwargs["name"]
        self.security_schemes = kwargs["security_schemes"]
        self.host = Host(**kwargs["host"])



from setuptools import setup, find_packages

setup(
    name="mcp-hive",
    version="0.2.0",
    packages=find_packages(),
    install_requires=[
        "rich==13.9.4",
        "python-dotenv==1.0.1",
        "langchain==0.3.15",
        "langchain-anthropic==0.3.4",
        "mcp==1.2.0",
    ],
    author="Ashish Mandal",
    author_email="mandal.ashish@codenation.co.in",
    description="Python SDK for MCP Hive - The hub for all your LLM tools",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/trilogy-group/mcp-hive",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)
